(*---------------------------------------------------------------------------*
  $Change$
  Copyright (c) 2006-2010, James H. Woodyatt
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Core I/O reactors. *)

(** {6 Overview}
    
    The {!Nx_poll} imperative event loop mechanism is lifted into a suite of
    I/O gadgets.  Compositions may use them to respond to process signals, loop
    idling conditions, timer events and file descriptor read/write/exception
    events discoverable through {!Unix.select}.
*)

(** {6 Event Types} *)

(** A process signal event. *)
type signal = [ `Signal of int ]

(** Loop idle event. *)
type idle = [ `Idle of Cf_tai64n.t ]

(** Timer event. *)
type time = [ `Time of Cf_tai64n.t ]

(** {6 Jacks} *)

(** The fix for process signal reactors. *)
type ('c, 'n) signaler_fix = ('c, 'n) Iom_gadget.fix
    constraint 'c = [> Iom_stream.flowcontrol ]
    constraint 'n = [> signal ]

(** The fix for loop idle reactors. *)
type ('c, 'n) idler_fix = ('c, 'n) Iom_gadget.fix
    constraint 'c = [> Iom_stream.flowcontrol ]
    constraint 'n = [> idle ]

(** The fix for timer event reactors. *)
type ('c, 'n) timer_fix = ('c, 'n) Iom_gadget.fix
    constraint 'c = [> Iom_stream.flowcontrol ]
    constraint 'n = [> time ]

(** A fix for reader reactors. *)
type reader_fix =
    (Iom_octet_stream.fragment, Iom_stream.flowcontrol, Iom_stream.failed)
    Iom_stream.ifix

(** A fix for writer reactors. *)
type writer_fix =
    (Iom_octet_stream.fragment, Iom_stream.stop, Iom_stream.flownotify)
    Iom_stream.ofix

(** {6 Functions}

    A simplified functional interface to the reactors.
*)

(** Evaluate [signaler ~n k] to obtain the signaler fix for a new reactor on
    the process signal [n] in the kernel context [k].
**)
val signaler: n:int -> Iom_gadget.kernel -> ('c, 'n) signaler_fix Iom_gadget.t

(** Evaluate [idler k] to obtain the idler fix for a new reactor on loop idle
    events in the kernel context [k].
*)
val idler: Iom_gadget.kernel -> ('c, 'n) idler_fix Iom_gadget.t

(** Evaluate [timer ?t0 ~dt k] to obtain the timer fix for a new reactor on
    timer events in the kernel context [k] that fire every [dt] seconds.
    Use the optional [t0] parameter to specify the time of the first event.
*)
val timer:
    ?t0:Cf_tai64n.t -> dt:float -> Iom_gadget.kernel ->
    ('c, 'n) timer_fix Iom_gadget.t

(** Evaluate [reader ~fd ~limits k] to obtain the reader fix for a new reactor
    on read events for the file descriptor [fd] in the kernel context [k] with
    [limits] on the buffering parameters.
*)
val reader:
    fd:Unix.file_descr -> limits:Iom_octet_stream.limits -> 
    Iom_gadget.kernel -> reader_fix Iom_gadget.t

(** Evaluate [writer ~fd k] to obtain the writer fix for a new reactor on
    write events for the file descriptor [fd] in the kernel context [k].
*)
val writer: fd:Unix.file_descr -> Iom_gadget.kernel -> writer_fix Iom_gadget.t

(** Use [retry_needed error] to test whether [error] is either [Unix.EAGAIN] or
    [Unix.EWOULDBLOCK].
*)
val retry_needed: Unix.error -> bool

(** {6 Classes}
    
    These classes are used to implement the simplified functions above.  They
    are provided to enable adaptations and extensions beyond the scope of the
    simplified interface.
*)

(** The core gadget for I/O reactors, comprising the basic control interface
    for stream gadgets and a type constraint on subtypes requiring the
    inheritance of the {!Nx_poll.event} class.  {b Note well:} these gadgets
    state objects are mutable, so care must be taken when writing derived
    classes to consider the side effects of mutation and not violate the
    requirements of the I/O gadget monad.
*)
class virtual ['control] core:
    'control Iom_gadget.rx ->
    object('self)
        inherit Iom_gadget.start
        inherit Iom_gadget.next
        constraint 'control = [> Iom_stream.stop ]
        constraint 'self = unit Iom_gadget.t #Nx_poll.event
        
        (** Set [true] when the gadget is started.  Because reactor state is
            mutable, attempts to start more than one gadget with the same
            initial state object raises [Failure].
        *)
        val mutable started_: bool
        
        (** Use [self#stop] to terminate the gadget.  In the base class, this
            just calls [Iom_gadget.abort].  Override to implement additional
            processing.
        *)
        method private stop: 'a. 'a Iom_gadget.t
        
        (** Override [self#control event] to implement control event processing
            for the reactor.  In the base class, all control events are
            ignored.
        *)
        method private control: 'control -> unit Iom_gadget.t
        
        (** Override [self#guard] to guard on all the receive wires in the
            gadget.  In the base class, just the control wire is guarded and
            the events are dispatched with [self#control].
        *)
        method private guard: unit Iom_gadget.guard
    end

(** A subclass of [core] that sends messages on a wire in response to polling
    events and flow control.
*)
class virtual ['control, 'source] source:
    k:Iom_gadget.kernel -> ('control, 'source) Iom_gadget.pad ->
    object('self)
        inherit ['control] core
        constraint 'control = [> Iom_stream.flowcontrol ]
        
        (** Set [true] when the flow controller is ready to receive events. *)
        val mutable source_: bool
        
        (** Define this method to produce a source message when the polling
            event [self#service] method is called.  Return [None] if no message
            is ready to send in response to this polling event.
        *)
        method virtual private source: 'source option
        
        (** Call [self#source], and if [source_] is still set afterward, then
            reload the polling event and send the message (if available) on the
            source wire.
        *)
        method private service: Nx_poll.t -> unit Iom_gadget.t Nx_poll.state
        
        (** Load the polling event and set [source_] to [true]. *)
        method private ready: unit Iom_gadget.t
        
        (** Unload the polling event and set [source_] to [false]. *)
        method private wait: unit Iom_gadget.t
    end

(** A stream gadget that responds to process signals. *)
class ['control, 'signal] signaler:
    k:Iom_gadget.kernel -> n:int -> ('control, 'signal) Iom_gadget.pad ->
    object
        inherit ['control, 'signal] source
        inherit [unit Iom_gadget.t] Nx_poll.signal
        constraint 'signal = [> signal ]
        
        (** Returns [Some (`Signal n)]. *)
        method private source: 'signal option
    end

(** A stream gadget that responds to loop idle events. *)
class ['control, 'idle] idler:
    k:Iom_gadget.kernel -> ('control, 'idle) Iom_gadget.pad ->
    object
        inherit ['control, 'idle] source
        inherit [unit Iom_gadget.t] Nx_poll.idle
        constraint 'idle = [> idle ]
        
        (** Returns [Some (`Idle epoch)] if [epoch_] is not [None], otherwise
            returns [None].
        *)
        method private source: 'idle option
    end

(** A stream gadget that responds to timer events. *)
class ['control, 'time] timer:
    k:Iom_gadget.kernel -> ?t0:Cf_tai64n.t -> dt:float ->
    ('control, 'time) Iom_gadget.pad ->
    object
        inherit ['control, 'time] source
        inherit [unit Iom_gadget.t] Nx_poll.time
        constraint 'time = [> time ]
        
        (** Returns [Some (`Time epoch_)]. *)
        method private source: 'time option
    end

(** A subclass of [source] that handles reading from a file descriptor when its
    polling event indicates it is ready.
*)
class virtual ['fragment, 'control, 'notify] reader_aux:
    k:Iom_gadget.kernel ->  ?rwx:Nx_poll.rwx -> fd:Unix.file_descr ->
    limits:Iom_octet_stream.limits ->
    ('fragment, 'control, 'notify) Iom_stream.ipad ->
    object('self)
        inherit ['control, 'fragment] source
        inherit [unit Iom_gadget.t] Nx_poll.file
        
        constraint 'fragment = #Iom_octet_stream.fragment
        constraint 'notify = [> Iom_stream.failed ]
        
        (** The octet buffer used for reading the file. *)
        val mutable buffer_: string
        
        (** The current offet into the buffer where octets will be stored. *)
        val mutable cursor_: int
        
        (** Indicates whether the end of the octet stream has been received. *)
        val mutable more_: Iom_stream.more
        
        (** The stack of buffers pending for delivery. *)
        val mutable stack_: (string * int * int) list
        
        (** The number of octets left to read before the high read limit. *)
        val mutable free_: int
        
        (** Define this method, called by [self#source] to construct an octet
            stream fragment object.
        *)
        method virtual private fragment: 'fragment
        
        (** Use [self#fail x] to abort the reactor with the exception [x].
            Sends [`Failed x] as a notification and calls [Iom_gadget.abort].
        *)
        method private fail: 'a. exn -> 'a Iom_gadget.t
        
        (** Invoked by [self#source] to read octets from the file descriptor
            into the buffer.  Defined in the base class as [(fun length ->
            Unix.read fd buffer_ cursor_ length)] and has the same semantics.
        *)
        method private octets: int -> int
        
        (** Enforces read limits and calls [self#octets] to read from the file
            descriptor, then calls itself recursively until the end of the
            octet stream is found, a failure exception arises, or the file
            descriptor blocks.
        *)
        method private source: 'fragment option
    end

(** A stream gadget that reads octet stream fragments with [Unix.read]. *)
class ['control, 'notify] reader:
    k:Iom_gadget.kernel -> ?rwx:Nx_poll.rwx -> fd:Unix.file_descr ->
    limits:Iom_octet_stream.limits ->
    (Iom_octet_stream.fragment, 'control, 'notify) Iom_stream.ipad ->
    object
        inherit [Iom_octet_stream.fragment, 'control, 'notify] reader_aux
        
        (** Returns a concrete octet stream fragment object. *)
        method private fragment: 'fragment
    end

(** A stream gadget that writes octet stream fragments with [Unix.write]. *)
class ['fragment, 'control, 'notify] writer:
    k:Iom_gadget.kernel -> fd:Unix.file_descr ->
    ('fragment, 'control, 'notify) Iom_stream.opad ->
    object('self)
        inherit ['control] core
        inherit [unit Iom_gadget.t] Nx_poll.file
        
        constraint 'notify = [> Iom_stream.flownotify ]
        constraint 'fragment = #Iom_octet_stream.fragment
        
        (** The queue of substrings pending for write. *)
        val mutable buffer_: Cf_message.substring Cf_deque.t
        
        (** Flag whether the end of the octet stream has been queued. *)
        val mutable more_: Iom_stream.more
        
        (** Use [self#fail x] to abort the reactor with the exception [x].
            Sends [`Failed x] as a notification and calls [Iom_gadget.abort].
        *)
        method private fail: 'a. exn -> 'a Iom_gadget.t
        
        (** Use [self#complete] to send the [`Wait] notification when the last
            octet has been written to the stream.  Derived classes may need to
            override.
        *)
        method private complete: unit Iom_gadget.t
        
        (** Service the polling event when the file descriptor is ready for
            writing.  Drains the pending buffer queue until empty, a failure
            exception arises or the file descriptor blocks.  Notifies [`Ready]
            if the queue drains completely.
        *)
        method private service: Nx_poll.t -> unit Iom_gadget.t Nx_poll.state
        
        (** Receive an octet stream fragment.  If the buffer queue is currently
            empty, the fragment is pushed into the buffer and [self#consume] is
            called to drain the queue.  If the queue is not emptied, then
            notifies [`Wait] and loads the polling event.
        *)
        method private receive: 'fragment -> unit Iom_gadget.t
        
        (** Invoked by [self#consume] to write octets to the file descriptor.
            Defined in the base class as identical to [Unix.write] and has the
            same semantics.
        *)
        method private octets: string -> int -> int -> int
        
        (** Drains the buffer queue.  Called in [self#service] when the polling
            event arises, and also called in [self#receive] when fragments
            arrive while the file descriptor is not blocked.
        *)
        method private consume: unit
    end

(*--- $File$ ---*)
