(*---------------------------------------------------------------------------*
  $Change$
  Copyright (c) 2006-2010, James H. Woodyatt
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** Core definitions for socket I/O gadgets. *)

(** {6 Overview}
    
    Sockets are full duplex communication mechanisms.  A single file descriptor
    comprises two independent reader and writer gadgets, and a unifying
    endpoint gadget for control of opening and closing semantics.  Classes and
    modules defined here are inherited and included by classes and modules
    defining specific forms of sockets.
*)

(** The core polymorphic variant type defining the state of a socket. *)
type ('ci, 'co, 'ni, 'no) state = [
        (** The socket has failed and is no longer usable. *)
        | `S_failed of exn
        
        (** The socket is established for simultaneous reading and writing. *)
        | `S_established of
            ('ci, 'ni) Iom_gadget.fix * ('co, 'no) Iom_gadget.fix
    ]
    constraint 'ci = [> Iom_stream.flowcontrol ]
    constraint 'co = [> Iom_stream.stop ]
    constraint 'ni = [> Iom_stream.failed ]
    constraint 'no = [> Iom_stream.flownotify ]

(** Use [inherit endpoint ~k ~s pad] to derive a subclass for the socket
    endpoint control semantics required by the type of socket [s].
*)
class virtual ['i, 'o, 'control, 'notify, 'state, 'af, 'st] endpoint :
    k:Iom_gadget.kernel -> s:('af, 'st) Nx_socket.t ->
    ('i, 'o, 'control, 'notify) Iom_stream.iopad ->
    object
        inherit Iom_gadget.start
        inherit Iom_gadget.next
        constraint 'control = [> Iom_stream.flowcontrol ]
        constraint 'notify = [> Iom_stream.flownotify ]
        constraint 'state = [> ('ci, 'co, 'ni, 'no) state ]
        
        (** The state of the socket. *)
        val state_: 'state
        
        (** Called by [self#establish] to start a receiver I/O gadget for the
            read side of the socket.
        *)
        method virtual private receiver:
            ('i, 'ci, 'ni) Iom_stream.ipad -> #Iom_gadget.start
        
        (** Called by [self#establish] to start a sender I/O gadget for the
            write side of the socket.
        *)
        method virtual private sender:
            ('o, 'co, 'no) Iom_stream.opad -> #Iom_gadget.start
        
        (** Calls [self#receiver] and [self#sender], starts the resulting
            receiver and sender gadgets, notifies [`Ready], and sets the socket
            into the [`Established] state.
        *)
        method private establish: 'state Iom_gadget.t
        
        (** Calls [self#establish].  Override in derived classes if connection
            oriented semantics are required.
        *)
        method private state0: 'state Iom_gadget.t
        
        (** Use [self#close] to close the socket and terminate the gadget.
            Derived classes may need to override.
        *)
        method private close: 'a. 'a Iom_gadget.t
        
        (** Use [self#reset] to shutdown the socket abortively, i.e. with
            [Unix.shutdown Unix.SHUTDOWN_ALL] and call [self#close].  Derived
            classes may need to override.
        *)
        method private reset: 'a. 'a Iom_gadget.t
        
        (** Use [self#fail x] to notify [`Failed x] and call [self#reset]. *)
        method private fail: 'a. exn -> 'a Iom_gadget.t
        
        (** Handle notifications from the receiver gadget.  Derived classes may
            need to override.  The base method catches [`Failed] notifications
            and stops the sender before calling [self#fail].  Unexpected
            notifications are ignored.
        *)
        method private inotify: 'ni -> unit Iom_gadget.t
        
        (** Handle notifications from the sender gadget.  Derived classes may
            need to override.  The base method forwards flow control
            notifications up the stack, and also catches [`Failed]
            notifications to stop the receiver before calling [self#fail].
            Unexpected notifications are ignored.
        *)
        method private onotify: 'no -> unit Iom_gadget.t
        
        (** Handle control messages.  Derived classes may need to override.
            The base method handles the [`Established] state and ignores
            all events in other states.  It forwards [`Stop] messages to both
            the receiver and the sender, and it forwards flow control messages
            to the receiver.
        *)
        method private control: 'control -> unit Iom_gadget.t
        
        (** Guards on notification events as specified by the socket state.  If
            the socket state is not [`Established] then no wires are guarded.
            Derived classes may need to override for handling guards in
            additional states.
        *)
        method private stateguard: unit Iom_gadget.guard
        
        (** Guards on the control wire, dispatching events with [self#control],
            and calls [self#stateguard] for the rest of the guard.
        *)
        method private guard: unit Iom_gadget.guard
    end

(** The signature of modules produced by the [Create] functor. *)
module type T = sig
    
    (** The socket module. *)
    module P: Nx_socket.P
    
    (** The type of socket descriptors. *)
    type t = (P.AF.tag, P.ST.tag) Nx_socket.t
    
    (** The type of socket addresses. *)
    type address = P.AF.address
    
    (** The "unspecified" address for the socket. *)
    val unspecified_address: address
    
    (** Use [socket ()] as a convenient constructor for socket descriptors. *)
    val create: unit -> t
    
    (** Use [socketpair ()] as a convenient constructor for the descriptors to
        a pair of connected sockets. *)
    val createpair: unit -> t * t
end

(** Use [Create(P)] to compose a module for including in modules for specific
    socket types that comprises the core definitions for socket I/O gadgets.
*)
module Create(P: Nx_socket.P): T with module P = P

(*--- $File$ ---*)
